// This is the code for the distributional model described in:
//    Feldman, N. H., Griffiths, T. L., Goldwater, S., & Morgan, J. L.  "A 
//    role for the developing lexicon in phonetic category acquisition."  
//    Psychological Review.
// Please cite that paper if you use this code.  You can contact the 
// corresponding author by emailing <nhf@umd.edu>.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version. This program is distributed in the 
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
// See the GNU General Public License for more details.  You should have 
// received a copy of the GNU General Public License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.
// 
// Copyright (C) 2013 Naomi Feldman

#include <stdio.h>
#include <assert.h>						// for safer programming
#include <math.h>							// needed for log, exp, and gamma functions
#include <gsl/gsl_matrix.h>		// gnu scientific library - matrices
#include <gsl/gsl_linalg.h>		// gnu scientific library - determinant
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>		// some vector functions
#include <stdlib.h>

// Constant
#define NUM_PHONES 6409      // total number of vowels in corpus

#define myrand() (double) (((unsigned long) randomMT()) / 4294967296.)
//#define myrand() (double) (rand()/(RAND_MAX+1.0))

// More Constants
const int MAX_ITER = 10000; // number of iterations
const int LIKELIHOOD_FREQ = 100; // likelihood is reported every LIKELIHOOD_FREQ iterations
const double ANNEALING_CONSTANT = 9.; // annealing constant
const int DIM = 2; // number of dimensions
const double PI = 3.1415926535897932384626433832795; // value of pi
const double LOGPI = 1.144729885849400174143; // value of log (pi)

// Type Definitions of Structures
typedef struct phoneme_struct {
	gsl_vector *mu_ptr; // vector corresponding to category mean
	gsl_matrix *varnu_ptr; // vector corresponding to category covariance
	double nu; // vector corresponding to category degrees of freedom (number of sounds assigned plus nu0)
	int nc; // number of times phoneme has been used, n_c
  int index; // index+1 of phoneme in the phoneme inventory, c_ind_ptr
	} phoneme;
typedef struct phone_struct {
	gsl_vector *formant_ptr; // vector corresponding to formant values of a particular acoustic token
	phoneme *z_ptr; // pointer to the phoneme to which it belongs
	} phone;

// Global Variables
phoneme *c_ind_ptr[NUM_PHONES+1]; // restaurant of (pointers to) phonemes
phone w[NUM_PHONES]; // list of phones in the corpus
gsl_vector *mu0_ptr; // prior on mean
gsl_matrix *varnu0_ptr; // prior on variance
double nu0; // prior on degrees of freedom
int num_tables_c; // number of tables being used in phoneme inventory
double beta;

// Function Declarations
void fileread (void);
void filewrite (int iteration, double ll);
void free_w (void);
double det (gsl_matrix *m_ptr, gsl_matrix *m_copy_ptr, gsl_permutation *p_ptr);
int random_sample (double probs[], double total_prob, int num_outcomes);
double t_likelihood (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, gsl_vector *mu_ptr, double nu, gsl_matrix *varnu_ptr, gsl_matrix *variance_ptr, gsl_matrix *matrix1_ptr, gsl_vector *diff_ptr, gsl_permutation *p_ptr);
void add_values (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, gsl_vector *mu_ptr, double *nu_ptr, gsl_matrix *varnu_ptr, gsl_vector *temp_ptr);
void subtract_values (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, gsl_vector *mu_ptr, double *nu_ptr, gsl_matrix *varnu_ptr, gsl_vector *temp_ptr);
phoneme *create_phoneme (void);
void destroy_phoneme (phoneme *phon_ptr);
void destroy_phoneme_index (phoneme *phon_ptr);
double joint_probability (gsl_matrix *matrix1_ptr, gsl_matrix *matrix2_ptr, gsl_vector *vector1_ptr, gsl_permutation *p_ptr);


/*******************************************************************************
void fileread(void) 

Reads values from a file.
********************************************************************************/

void fileread (void) 
{
  int i;
  FILE *vowel_file_ptr; // a file that contains a list of acoustic values

  // open the datafile
  vowel_file_ptr = fopen("Simulation3-vowel.txt", "r");
  
  i = 0; // counter for which phone we're reading in
  
  while (1) {
    
    w[i].formant_ptr = gsl_vector_alloc (DIM);
    gsl_vector_fscanf (vowel_file_ptr, w[i].formant_ptr);
    
    i++;
    
    if (i == NUM_PHONES) {break;}

  }
  
  fclose (vowel_file_ptr);
}


/*******************************************************************************
void filewrite(int iteration, double ll) 

Writes results to a file.
- iteration is the iteration number, used in naming the files
********************************************************************************/

void filewrite (int iteration, double ll)
{
  FILE *z_file_ptr; // file with word assignments to lexical items
  FILE *c_file_ptr; // file with phoneme parameters
  FILE *ll_file_ptr; // file that has the joint probability
  
  char z[50], c[50], lik[50]; // sprintf needs variables to write to
  sprintf (z, "z_%i_%i.txt", (int) (beta*10), iteration);
  sprintf (c, "c_%i_%i.txt", (int) (beta*10), iteration);
  sprintf (lik, "ll_%i_%i.txt", (int) (beta*10), iteration);
  
  z_file_ptr = fopen (z, "w");
  c_file_ptr = fopen (c, "w");
  ll_file_ptr = fopen (lik, "w");
  
  // print results to these files
  int i;
  // print word assignments vector z
  for (i = 0; i < NUM_PHONES; i++) {
    fprintf (z_file_ptr, "%i\n", w[i].z_ptr->index);
  }
  // print phonetic category parameters c
  for (i = 0; i < num_tables_c; i++) {
    fprintf (c_file_ptr, "%g %g\n%g %g %g %g\n%g\n", gsl_vector_get(c_ind_ptr[i]->mu_ptr,0), 
             gsl_vector_get(c_ind_ptr[i]->mu_ptr,1), gsl_matrix_get(c_ind_ptr[i]->varnu_ptr,0,0),
             gsl_matrix_get(c_ind_ptr[i]->varnu_ptr,0,1), gsl_matrix_get(c_ind_ptr[i]->varnu_ptr,1,0),
             gsl_matrix_get(c_ind_ptr[i]->varnu_ptr,1,1), c_ind_ptr[i]->nu);
  }
  fprintf (ll_file_ptr, "%g\n", ll);
  
  fclose (z_file_ptr);
  fclose (c_file_ptr);
  fclose (ll_file_ptr);
}

/*******************************************************************************
void free_w(void) 

Frees the memory associated with the corpus.
********************************************************************************/

void free_w (void) 
{
  int i;
  
  for (i = 0; i < NUM_PHONES; i++) {
    // free memory associated with formant vectors
    gsl_vector_free (w[i].formant_ptr);
  }
  
}


/*******************************************************************************
double det(gsl_matrix *m_ptr)

Returns the determinant of a matrix m_ptr, without overwriting the contents of m_ptr.
********************************************************************************/

double det (gsl_matrix *m_ptr, gsl_matrix *m_copy_ptr, gsl_permutation *p_ptr)
{
  int signum;
	gsl_matrix_memcpy (m_copy_ptr, m_ptr);
	gsl_linalg_LU_decomp (m_copy_ptr, p_ptr, &signum);  // LU decomposition
	double d = gsl_linalg_LU_det (m_copy_ptr, signum); // use the diagonal to compute the determinant
	return d;
}


/*******************************************************************************
int random_sample(probs, totprob, num_outcomes)

Returns a random sample from a discrete probability distribution.
- probs is a pointer to an array of counts (proportional to probabilities)
- totprob is the sum of everything in probs -- the proportionality constant
- num_outcomes is the number of discrete bins
********************************************************************************/

int random_sample (double probs[], double total_prob, int num_outcomes)
{
	int i = 0; // bin number
	double r = myrand() * total_prob; // pick a random number between 0 and totprob
	double cutoff = *probs; // maximum-number cutoff for r to fall into bin i
  //printf ("Number: %g; Totprob: %g\n", r, total_prob);
	while (r > cutoff) {
		i++;
		assert (i < num_outcomes); // make sure we haven't run out of options
		cutoff += *(probs+i); // pointer arithmetic
  }
  assert (probs[i] >= 0);
	return i; // returns a number between 0 and num_outcomes-1
}

/*******************************************************************************
double t_likelihood(wkj, ss, n, mu, nu, varnu)

Returns the log likelihood of a datapoint belonging to a category that contains 
a given set of other points.
The first 3 parameters describe the data:
- wkj is the mean of the query points
- ss is the sum of squares from the data, i.e. sum((WIJ-WKJ)^2)
- n is the number of query points
The last 3 parameters describe the category:
- mu is the mean of the likelihood
- nu is the degrees of freedom of the likelihood
- varnu is the variance * degrees of freedom of the likelihood
********************************************************************************/

double t_likelihood (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, 
										 gsl_vector *mu_ptr, double nu, gsl_matrix *varnu_ptr, 
                     gsl_matrix *variance_ptr, gsl_matrix *matrix1_ptr, gsl_vector *diff_ptr, 
                     gsl_permutation *p_ptr)
{
	double p = 0; // log probability
	int i;
	gsl_matrix_memcpy (variance_ptr, varnu_ptr); // variance_ptr holds varnu
	gsl_matrix_add (variance_ptr, ss_ptr); // variance_ptr holds varnu+ss
	gsl_vector_memcpy (diff_ptr, wkj_ptr); // diff_ptr holds wkj
	gsl_vector_sub (diff_ptr, mu_ptr); // diff_ptr holds wkj-mu
	gsl_blas_dger ((n * nu)/(n + nu), diff_ptr, diff_ptr, variance_ptr);  // variance_ptr holds entire sums-of-squares term

	// compute the normalization constant
	for (i = 0; i < DIM; i++) {
		p += lgamma ((nu + n - (double) i) / 2.) - lgamma ((nu - (double) i) / 2.);
	}
	// compute the rest of the log probability
	p += (nu / 2) * log (det (varnu_ptr, matrix1_ptr, p_ptr)) - ((double) DIM * n/2) * LOGPI - ((double) DIM / 2) * log ((n + nu) / nu )
		- ((nu + n) / 2) * log (det (variance_ptr, matrix1_ptr, p_ptr));
		
	return p;
}


/*******************************************************************************
void add_values(wkj, ss, n, *mu, *nu, *varnu)

Adds the values from a set of datapoints to the parameters of a category.
The first 3 parameters describe the data:
- wkj is the mean of the query points
- ss is the sum of squares from the data, i.e. sum((WIJ-WKJ)^2)
- n is the number of query points
The last 3 parameters describe the category:
- mu is the mean of the likelihood
- nu is the degrees of freedom of the likelihood
- varnu is the variance * degrees of freedom of the likelihood
********************************************************************************/

void add_values (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, 
							   gsl_vector *mu_ptr, double *nu_ptr, gsl_matrix *varnu_ptr, gsl_vector *temp_ptr)
{
	gsl_matrix_add(varnu_ptr, ss_ptr);
	gsl_vector_memcpy (temp_ptr, wkj_ptr); // temp_ptr holds wkj
	gsl_vector_sub (temp_ptr, mu_ptr); // temp_ptr holds wkj-mu
	gsl_blas_dger ((*nu_ptr * n) / (*nu_ptr + n), temp_ptr, temp_ptr, varnu_ptr);
	gsl_vector_scale (mu_ptr, (*nu_ptr / (*nu_ptr + n)));
	gsl_vector_memcpy (temp_ptr, wkj_ptr); // temp_ptr holds wkj
	gsl_vector_scale (temp_ptr,  n / (*nu_ptr + n) ); // temp_ptr holds a scaled version of wkj
	gsl_vector_add (mu_ptr, temp_ptr);
	*nu_ptr += n;
	// make sure the matrix is symmetric (this is for 2x2 covariance matrices -- need to generalize to more dimensions)
	if (gsl_matrix_get (varnu_ptr, 0, 1) != gsl_matrix_get (varnu_ptr, 1, 0)) {
		gsl_matrix_set (varnu_ptr, 0, 1, gsl_matrix_get (varnu_ptr, 1, 0));
	}
}

/*******************************************************************************
void subtract_values(wkj, ss, n, *mu, *nu, *varnu)

Subtracts the values from a set of datapoints to the parameters of a category.
The first 3 parameters describe the data:
- wkj is the mean of the query points
- ss is the sum of squares from the data, i.e. sum((WIJ-WKJ)^2)
- n is the number of query points
The last 3 parameters describe the category:
- mu is the mean of the likelihood
- nu is the degrees of freedom of the likelihood
- varnu is the variance * degrees of freedom of the likelihood
********************************************************************************/

void subtract_values (gsl_vector *wkj_ptr, gsl_matrix *ss_ptr, double n, 
                      gsl_vector *mu_ptr, double *nu_ptr, gsl_matrix *varnu_ptr, gsl_vector *temp_ptr)
{
	*nu_ptr -= n;
	gsl_vector_memcpy (temp_ptr, wkj_ptr); // temp_ptr holds wkj
  gsl_vector_scale (temp_ptr,  n / (*nu_ptr + n) ); // temp_ptr holds a scaled version of wkj
  gsl_vector_sub (mu_ptr, temp_ptr); // subtract scaled wkj
  gsl_vector_scale (mu_ptr, (*nu_ptr + n) / *nu_ptr ); // multiply by appropriate constant
	gsl_vector_memcpy (temp_ptr, wkj_ptr); // temp_ptr holds wkj
	gsl_vector_sub (temp_ptr, mu_ptr); // temp_ptr holds wkj-mu
  gsl_matrix_sub (varnu_ptr, ss_ptr);
  gsl_blas_dger (-1. * (*nu_ptr * n) / (*nu_ptr + n), temp_ptr, temp_ptr, varnu_ptr);

	// make sure the matrix is symmetric (this is for 2x2 covariance matrices -- need to generalize to more dimensions)
	if (gsl_matrix_get (varnu_ptr, 0, 1) != gsl_matrix_get (varnu_ptr, 1, 0)) {
		gsl_matrix_set (varnu_ptr, 0, 1, gsl_matrix_get (varnu_ptr, 1, 0));
	}
}


/*******************************************************************************
phoneme *create_phoneme(void)

Allocates memory for a new phoneme, and initializes its values.
********************************************************************************/

phoneme *create_phoneme (void)
{
	phoneme *phon_ptr = malloc(sizeof(phoneme));
	if (phon_ptr == NULL)	{ fprintf(stderr, "out of memory\n"); exit(EXIT_FAILURE); }
	phon_ptr -> mu_ptr = gsl_vector_alloc (DIM);
  gsl_vector_memcpy (phon_ptr -> mu_ptr, mu0_ptr);
	phon_ptr -> varnu_ptr = gsl_matrix_alloc (DIM, DIM);
  gsl_matrix_memcpy (phon_ptr -> varnu_ptr, varnu0_ptr);
	phon_ptr -> nu = nu0;
	phon_ptr -> nc = 0;
  phon_ptr -> index = 0; // for bookkeeping
	return phon_ptr;
}

/*******************************************************************************
void destroy_phoneme(*phon_ptr)

Frees the memory associated with a phoneme.
- phon_ptr is a pointer to the phoneme that needs to be freed
********************************************************************************/

void destroy_phoneme (phoneme *phon_ptr)
{
  gsl_vector_free (phon_ptr -> mu_ptr);
  phon_ptr -> mu_ptr = NULL;
  gsl_matrix_free (phon_ptr -> varnu_ptr);
  phon_ptr -> varnu_ptr = NULL;
	free (phon_ptr);
}


/*******************************************************************************
void destroy_phoneme_index(*phon_ptr)

Deletes a phoneme from the phoneme inventory, but doesn't free the memory.
- phon_ptr is a pointer to the phoneme whose index needs to be deleted
********************************************************************************/

void destroy_phoneme_index (phoneme *phon_ptr)
{
	// find the table number to which this phoneme corresponds
	int i = 0, j;
	while (c_ind_ptr[i] != phon_ptr) {
		i++;
		assert (c_ind_ptr[i] != NULL);
	}
	for (j = i; c_ind_ptr[j] != NULL; j++) { // null value marks the end of the restaurant
		c_ind_ptr[j] = c_ind_ptr[j+1]; // transpose the table identities by one
    if (c_ind_ptr[j] != NULL) {
      c_ind_ptr[j]->index--; // decrease index number on that phoneme
      assert (c_ind_ptr[j]->index > 0);
    }
	}
}


/*******************************************************************************
double joint_probability(void)

Computes the joint probability of the data and category assignments by factorizing
the joint probability of the data:
p(w,z|beta,mu0,nu0,var0) = p(w|z,mu0,nu0,var0) p(z|beta)
********************************************************************************/

double joint_probability (gsl_matrix *matrix1_ptr, gsl_matrix *matrix2_ptr, gsl_vector *vector1_ptr, gsl_permutation *p_ptr)
{
  int i, j;
  double total_n = 0;
  double logprob = (double) num_tables_c * log (beta); // beta factor in the numerator
  // p(z|beta)
  // The numerator is beta * 1 * 2 * ... * Nc-1 for each category
  // The denominator is beta * (1+beta) * (2+beta) * ... * (NC-1+beta)
  // compute the rest of the numerator
  for (i = 0; i < num_tables_c; i++) {
    logprob += lgamma ((double) c_ind_ptr[i]->nc);
    total_n += (double) c_ind_ptr[i]->nc;
  }
  // compute the denominator
  for (i = 0; i < total_n; i++) {
    logprob -= log ((double) i + beta);
  }
  
  // p(w|z,mu0,nu0,var0)
  // Probability of the acoustic values given their phoneme assignments
  gsl_vector *vzeros_ptr = gsl_vector_calloc (DIM);
  gsl_matrix *mzeros_ptr = gsl_matrix_calloc (DIM, DIM);
  gsl_vector *wkj_ptr = gsl_vector_alloc (DIM);
  gsl_matrix *ss_ptr = gsl_matrix_alloc (DIM, DIM);
  for (i = 0; i < num_tables_c; i++) { // loop through all the phonemes
    // Reset the mean, covariance, and df of the phoneme
    gsl_vector_memcpy (c_ind_ptr[i]->mu_ptr, mu0_ptr);
    gsl_matrix_memcpy (c_ind_ptr[i]->varnu_ptr, varnu0_ptr);
    c_ind_ptr[i]->nu = nu0;
    int n = 0; // count the number of acoustic values assigned to the phoneme
    // Initialize mean and ss variables to zero
    gsl_vector_memcpy (wkj_ptr, vzeros_ptr);
    gsl_matrix_memcpy (ss_ptr, mzeros_ptr);
    // Find instances of that phoneme
    for (j = 0; j < NUM_PHONES; j++) { // search through all words
      if (w[j].z_ptr == c_ind_ptr[i]) {
        // add the value to the running mean/cov/df counts for the phoneme
        gsl_vector_add (wkj_ptr, w[j].formant_ptr);
        gsl_blas_dger (1., w[j].formant_ptr, w[j].formant_ptr, ss_ptr);
        n++;
      }
    }
    // Correct the mean and sum-of-squares terms
    gsl_vector_scale (wkj_ptr, 1./(double) n); // normalize the mean, E[X]
    gsl_blas_dger (-1.*(double) n, wkj_ptr, wkj_ptr, ss_ptr); // n*E[X^2]-n*E[X]^2
    // Multiply by the likelihood of these points
    logprob += t_likelihood (wkj_ptr, ss_ptr, (double) n, c_ind_ptr[i]->mu_ptr, c_ind_ptr[i]->nu, c_ind_ptr[i]->varnu_ptr, 
                             matrix1_ptr, matrix2_ptr, vector1_ptr, p_ptr);
    // Compute the new mean and covariance counts for the phoneme
    add_values (wkj_ptr, ss_ptr, (double) n, c_ind_ptr[i]->mu_ptr, &c_ind_ptr[i]->nu, c_ind_ptr[i]->varnu_ptr, vector1_ptr);
  }
  
  gsl_vector_free (wkj_ptr);
  gsl_matrix_free (ss_ptr);
  
  gsl_vector_free (vzeros_ptr);
  gsl_matrix_free (mzeros_ptr);
  
  return logprob;

}

/*******************************************************************************
int main()

This main program has two stages associated with it:
(1) Initialization
		- Set priors on phonetic category parameters
		- Initialize all words to different lexical items
		- Initialize all phoneme slots in all lexical items to the same phoneme
(2) Re-sample Z (phoneme assignments)
		- Compute likelihood for all phonemes
		- Sample a phoneme
********************************************************************************/


int main()
{

  seedMT (time(NULL));  
//	unsigned int iseed = (unsigned int)time(NULL);
//  srand (iseed);
	
	int i, k;
    
  // To print the first few random numbers
  for (i=0; i<0; i++) {
    double r = myrand();
    printf ("%g\n",r);
  }

	// Read data from a file
	printf ("Reading file...\n");
  fileread();
  
	// Local Variables
  gsl_matrix *zeros = gsl_matrix_calloc (DIM,DIM); // zero covariance matrix for bookkeeping
  
  // Bookkeeping matrices that get passed to functions and overwritten each time
  gsl_matrix *matrix1_ptr = gsl_matrix_alloc (DIM, DIM);
  gsl_matrix *matrix2_ptr = gsl_matrix_alloc (DIM, DIM);
  gsl_vector *vector1_ptr = gsl_vector_alloc (DIM);
  gsl_permutation *p_ptr = gsl_permutation_alloc (DIM);
  
	/* INITIALIZE */
  
  printf ("Setting constants...\n");
	/* Priors on phonetic category parameters, a.k.a. Even More Constants */
	mu0_ptr = gsl_vector_alloc (DIM); // prior on mean
	gsl_vector_set (mu0_ptr, 0, 500.); // set prior to the correct value
	gsl_vector_set (mu0_ptr, 1, 1500.);
	varnu0_ptr = gsl_matrix_alloc (DIM,DIM); // prior on variance
	gsl_matrix_set_identity (varnu0_ptr); // set to the identity matrix
  nu0 = (double) DIM - 1. + .001; // prior on degrees of freedom
  /* Done with prior on phonetic category parameters */

  for (beta = 10; beta < 100; beta *= 10) {

    /* INITIALIZATION */
    gsl_vector *wkj_ptr = gsl_vector_calloc (DIM); // initialize wkj, set elements to zero
    gsl_matrix *ss_ptr = gsl_matrix_calloc (DIM, DIM); // initialize ss
    num_tables_c = 0;  

    double anneal = (log (2.)) / ANNEALING_CONSTANT;

    // loop through all phones in the corpus
    for (i = 0; i < NUM_PHONES; i++) {
      double probs[num_tables_c+1];
      double *max_prob_ptr = probs; // keep track of the highest probability, to subtract it off
      for (k = 0; k < num_tables_c; k++) {
        // compute log likelihood for all phonemes
        // and multiply by something proportional to the prior
        probs[k] = t_likelihood (w[i].formant_ptr, zeros, 1., c_ind_ptr[k]->mu_ptr,
                                c_ind_ptr[k]->nu, c_ind_ptr[k]->varnu_ptr, matrix1_ptr, matrix2_ptr,
                                vector1_ptr, p_ptr);
        probs[k] += log ((double) c_ind_ptr[k]->nc);
        if (probs[k] > *max_prob_ptr) {
          max_prob_ptr = probs + k; // pointer arithmetic for this phonetic category probability
        }
      }
      // log probability of a new table
      probs[num_tables_c] = t_likelihood(w[i].formant_ptr, zeros, 1., mu0_ptr, nu0,
                                         varnu0_ptr, matrix1_ptr, matrix2_ptr, vector1_ptr, p_ptr) 
                                         + log (beta);
      if (probs[num_tables_c] > *max_prob_ptr) {
        max_prob_ptr = probs + num_tables_c; // pointer arithmetic for this phonetic category probability
      }
      
      double max = *max_prob_ptr; // copy the value to a safe place
      max_prob_ptr = NULL; // for bookkeeping, and to avoid errors
      double totprob = 0.;
      for (k = 0; k <= num_tables_c; k++) {
        probs[k] = exp (anneal * (probs[k]-max));
        totprob += probs[k]; // normalization constant
      }
      
      // Sample a phonetic category
      int assign = random_sample (probs, totprob, num_tables_c+1);
      if (assign == num_tables_c) {
        //printf ("Creating phoneme %i\n", num_tables_c+1);
        c_ind_ptr[num_tables_c] = create_phoneme();
        c_ind_ptr[num_tables_c]->index = num_tables_c+1;
        num_tables_c++;
        c_ind_ptr[num_tables_c] = NULL;
      }
      w[i].z_ptr = c_ind_ptr[assign];
      add_values (w[i].formant_ptr, zeros, 1., w[i].z_ptr->mu_ptr, 
                  &w[i].z_ptr->nu, w[i].z_ptr->varnu_ptr, 
                  vector1_ptr); // add values to new phoneme
      w[i].z_ptr->nc++; // increment the phoneme count
  
    }
    
    /* END INITIALIZATION */
    
    
    
    int iteration;
    
    for (iteration = 0; iteration < MAX_ITER; iteration++) {
      
      printf ("Iteration %i: \n", iteration);
      
      //double anneal = 1;
      double anneal = (log ((double) iteration+2.)) / ANNEALING_CONSTANT;
      if (anneal > 1.) {
        anneal = 1.;
      }
      
      
      /* SWEEP TO RE-SAMPLE Z */
      
      assert (c_ind_ptr[num_tables_c] == NULL);
      
      printf ("Sampling Z...\n");
      // loop through all phonemes
      for (i = 0; i < NUM_PHONES; i++) {
        // Decrement the count of the current phoneme
        w[i].z_ptr->nc--;
        // Delete the values from the phoneme
        subtract_values (w[i].formant_ptr, zeros, 1., w[i].z_ptr->mu_ptr, 
                         &w[i].z_ptr->nu, w[i].z_ptr->varnu_ptr, vector1_ptr);
        // If the phoneme is no longer used, delete the table
        if (w[i].z_ptr->nc == 0) {
          //printf ("Destroying phoneme %i\n", l_ind_ptr[i]->vowel_ptr[j].phon_ptr->index);
          destroy_phoneme_index (w[i].z_ptr); // delete it from the restaurant
          destroy_phoneme (w[i].z_ptr); // and free the memory
          num_tables_c--;
        }
        w[i].z_ptr = NULL;
        
        double probs[num_tables_c+1];
        double *max_prob_ptr = probs; // keep track of the highest probability, to subtract it off
        for (k = 0; k < num_tables_c; k++) {
          // compute log likelihood for all phonemes
          // and multiply by something proportional to the prior
          probs[k] = t_likelihood (w[i].formant_ptr, zeros, 1., c_ind_ptr[k]->mu_ptr,
                                  c_ind_ptr[k]->nu, c_ind_ptr[k]->varnu_ptr, matrix1_ptr, matrix2_ptr,
                                  vector1_ptr, p_ptr);
          probs[k] += log ((double) c_ind_ptr[k]->nc);
          if (probs[k] > *max_prob_ptr) {
            max_prob_ptr = probs + k; // pointer arithmetic for this phonetic category probability
          }
        }
        // log probability of a new table
        probs[num_tables_c] = t_likelihood(w[i].formant_ptr, zeros, 1., mu0_ptr, nu0,
                                           varnu0_ptr, matrix1_ptr, matrix2_ptr, vector1_ptr, p_ptr) 
                                           + log (beta);
        if (probs[num_tables_c] > *max_prob_ptr) {
          max_prob_ptr = probs + num_tables_c; // pointer arithmetic for this phonetic category probability
        }
        
        double max = *max_prob_ptr; // copy the value to a safe place
        max_prob_ptr = NULL; // for bookkeeping, and to avoid errors
        double totprob = 0.;
        for (k = 0; k <= num_tables_c; k++) {
          probs[k] = exp (anneal * (probs[k]-max));
          totprob += probs[k]; // normalization constant
        }
        
        // Sample a phonetic category
        int assign = random_sample (probs, totprob, num_tables_c+1);
        if (assign == num_tables_c) {
          //printf ("Creating phoneme %i\n", num_tables_c+1);
          c_ind_ptr[num_tables_c] = create_phoneme();
          c_ind_ptr[num_tables_c]->index = num_tables_c+1;
          num_tables_c++;
          c_ind_ptr[num_tables_c] = NULL;
        }
        w[i].z_ptr = c_ind_ptr[assign];
        add_values (w[i].formant_ptr, zeros, 1., w[i].z_ptr->mu_ptr, 
                    &w[i].z_ptr->nu, w[i].z_ptr->varnu_ptr, 
                    vector1_ptr); // add values to new phoneme
        w[i].z_ptr->nc++; // increment the phoneme count
      }
      
      /* END SWEEP TO RE-SAMPLE Z */
            
      if ((iteration+1) % LIKELIHOOD_FREQ == 0) {
        // Refresh parameters, and compute joint probability
        double ll = joint_probability(matrix1_ptr, matrix2_ptr, vector1_ptr, p_ptr);
        printf ("Iteration %i: %g\n", iteration, ll);
        filewrite(iteration, ll);
      }
      printf ("%i phonemes\n", num_tables_c);
    }
    
    
    /* CLEAN-UP */
    
    while (c_ind_ptr[0] != NULL) {
      phoneme *temp_ptr = c_ind_ptr[0];
      destroy_phoneme_index (temp_ptr);
      destroy_phoneme (temp_ptr);
    }
    
    gsl_vector_free (wkj_ptr);
    gsl_matrix_free (ss_ptr);
    

  
  }
  
  
  free_w();

  gsl_matrix_free (matrix1_ptr);
  gsl_matrix_free (matrix2_ptr);
  gsl_vector_free (vector1_ptr);
  gsl_permutation_free (p_ptr);

  gsl_vector_free (mu0_ptr);
  gsl_matrix_free (varnu0_ptr);
  gsl_matrix_free (zeros);
  
  /* END CLEAN-UP */

  return 0; // tell the user the program was successful

}